// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO20/16AO20_Linux_2.x.x.x_GSC_DN/utils/util_fsamp_ao.c $
// $Rev: 52760 $
// $Date: 2023-04-05 15:06:22 -0500 (Wed, 05 Apr 2023) $

// 16AO20: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	ABS(v)					(((v) < 0) ? (-(v)) : (v))
#define	LIMIT_MAX(v,max)		(((v) > (max)) ? (max) : (v))
#define	LIMIT_MIN(v,min)		(((v) < (min)) ? (min) : (v))
#define	LIMIT_RANGE(v,min,max)	(LIMIT_MIN((LIMIT_MAX((v),(max))),(min)))



// data types *****************************************************************

typedef struct
{
	s32			fsamp_want;		// value provided
	s32			fref;			// value provided/updated

	s32			fsamp_max;		// queried value
	s32			fsamp_min;		// queried value
	s32			nclk_max;		// queried value
	s32			nclk_min;		// queried value
	s32			nrate_max;		// queried value
	s32			nrate_min;		// queried value

	double		delta;			// value updated
	double		fsamp_got;		// value computed
	s32			nclk;			// value computed
	s32			nrate;			// value computed
	s32			osc_src;		// value computed
	const char*	osc_src_name;	// value computed
} fsamp_t;



//*****************************************************************************
static void _fsamp_alternate_compute(fsamp_t* data)
{
	double	adj;
	double	delta;
	double	fsamp_got;
	s32		nclk;
	s32		nrate;
	s32		nrate_ll;
	s32		nrate_ul;

	data->fref			= 16000000;
	data->osc_src		= AO20_CLOCK_OSC_SRC_ALT;
	data->osc_src_name	= "Alternate";

	for (nclk = data->nclk_min; nclk <= data->nclk_max; nclk++)
	{
		adj			= (double) data->fref * (1 + ((double) nclk / 511));
		nrate		= (s32) (adj / data->fsamp_want);
		nrate_ll	= nrate - 1;
		nrate_ul	= nrate + 1;
		nrate_ll	= (nrate_ll < data->nrate_min) ? data->nrate_min : nrate_ll;
		nrate_ll	= (nrate_ll > data->nrate_max) ? data->nrate_max : nrate_ll;
		nrate_ul	= (nrate_ul > data->nrate_max) ? data->nrate_max : nrate_ul;
		nrate_ul	= (nrate_ul < data->nrate_min) ? data->nrate_min : nrate_ul;

		for (nrate = nrate_ll; nrate <= nrate_ul; nrate++)
		{
			fsamp_got	= ((double) adj) / nrate;
			delta		= fsamp_got - data->fsamp_want;

			if ((ABS(delta)) < (ABS(data->delta)))
			{
				data->delta		= delta;
				data->fsamp_got	= fsamp_got;
				data->nclk		= nclk;
				data->nrate		= nrate;
			}
		}
	}
}



//*****************************************************************************
static void _fsamp_primary_compute(fsamp_t* data)
{
	double	delta;
	double	fsamp_got;
	s32		nrate;
	s32		nrate_ll;
	s32		nrate_ul;

	data->osc_src		= AO20_CLOCK_OSC_SRC_PRI;
	data->osc_src_name	= "Primary";

	nrate		= (long) (data->fref / data->fsamp_want);
	nrate_ll	= nrate - 1;
	nrate_ul	= nrate + 1;
	nrate_ll	= (nrate_ll < data->nrate_min) ? data->nrate_min : nrate_ll;
	nrate_ll	= (nrate_ll > data->nrate_max) ? data->nrate_max : nrate_ll;
	nrate_ul	= (nrate_ul > data->nrate_max) ? data->nrate_max : nrate_ul;
	nrate_ul	= (nrate_ul < data->nrate_min) ? data->nrate_min : nrate_ul;

	for (nrate = nrate_ll; nrate <= nrate_ul; nrate++)
	{
		fsamp_got	= ((double) data->fref) / nrate;
		delta		= fsamp_got - data->fsamp_want;

		if ((ABS(delta)) < (ABS(data->delta)))
		{
			data->delta		= delta;
			data->fsamp_got	= fsamp_got;
			data->nrate		= nrate;
		}
	}
}



/******************************************************************************
*
*	Function:	ao20_fsamp_ao_compute
*
*	Purpose:
*
*		Determine the best configuration for the specified sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	The desired sample rate.
*
*		osc_src	The oscillator source is put here. MUST NOT BE NULL!
*
*		nclk	The Nclk value is put here. MUST NOT BE NULL!
*
*		nrate	The Nrate value is put here. MUST NOT BE NULL!
*
*		rate	Report the sample rate achieved here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int ao20_fsamp_ao_compute(int fd, int index, int verbose, s32 fsamp, s32* osc_src, s32* nclk, s32* nrate, double* rate)
{
	fsamp_t		data;
	fsamp_t		data_alt;
	fsamp_t		data_pri;
	int			errs		= 0;

	if (verbose)
	{
		gsc_label_index("AO Sample Rate", index);
		gsc_label_long_comma(fsamp);
		printf(" S/S");
	}

	memset(&data, 0, sizeof(data));
	errs	+= ao20_query(fd, -1, 0, AO20_QUERY_FREF_DEFAULT, &data.fref);
	errs	+= ao20_query(fd, -1, 0, AO20_QUERY_FSAMP_MAX, &data.fsamp_max);
	errs	+= ao20_query(fd, -1, 0, AO20_QUERY_FSAMP_MIN, &data.fsamp_min);
	errs	+= ao20_query(fd, -1, 0, AO20_QUERY_NCLK_MAX, &data.nclk_max);
	errs	+= ao20_query(fd, -1, 0, AO20_QUERY_NCLK_MIN, &data.nclk_min);
	errs	+= ao20_query(fd, -1, 0, AO20_QUERY_NRATE_MAX, &data.nrate_max);
	errs	+= ao20_query(fd, -1, 0, AO20_QUERY_NRATE_MIN, &data.nrate_min);

	data.fsamp_want	= (double) LIMIT_RANGE(fsamp, data.fsamp_min, data.fsamp_max);
	data.delta		= FLT_MAX;
	data.nclk		= -1;
	data.nrate		= -1;
	data.osc_src	= -1;

	data_alt	= data;
	data_pri	= data;

	_fsamp_alternate_compute(&data_alt);
	_fsamp_primary_compute(&data_pri);

	if ((ABS(data_alt.delta)) < (ABS(data_pri.delta)))
		data	= data_alt;
	else
		data	= data_pri;

	// Report the results.

	osc_src[0]	= data.osc_src;
	nrate[0]	= data.nrate;
	nclk[0]		= data.nclk;

	if (rate)
		rate[0]	= data.fsamp_got;

	if (verbose)
	{
		printf("%s  (", errs ? " FAIL <---" : "");
		printf("%s", data.osc_src_name);
		printf(", Fref ");
		gsc_label_long_comma(data.fref);
		printf(", Nrate ");
		gsc_label_long_comma(data.nrate);

		if (data.osc_src == AO20_CLOCK_OSC_SRC_ALT)
		{
			printf(", Nclk ");
			gsc_label_long_comma(data.nclk);
		}

		printf(", Fsamp %.3f", data.fsamp_got);
		printf(")\n");
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	ao20_fsamp_get
*
*	Purpose:
*
*		Determine the sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		title	Show a title?
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int ao20_fsamp_get(int fd, int title, s32* get)
{
	s32	clk_src;
	int	errs		= 0;
	s32	fref;
	s32	nclk;
	s32	nrate;
	s32	osc_src;
	s32	set			= 0;

	if (title)
		gsc_label("Sample Rate");

	errs	+= ao20_clock_osc_src(fd, -1, 0, -1, &osc_src);
	errs	+= ao20_fref_compute(fd, -1, 0, &fref);
	errs	+= ao20_clock_src(fd, -1, 0, -1, &clk_src);
	errs	+= ao20_nclk(fd, -1, 0, -1, &nclk);
	errs	+= ao20_nrate(fd, -1, 0, -1, &nrate);

	if (clk_src == AO20_CLOCK_SRC_EXT_SW)
	{
		set	= 0;
		printf("Unknown  (Source = External/Software)\n");
	}
	else if (osc_src == AO20_CLOCK_OSC_SRC_PRI)
	{
		set	= fref / nrate;
		gsc_label_long_comma(set);
		printf(" S/S  (Source = Internal, Pri Ref, Nrate = ");
		gsc_label_long_comma(nrate);
		printf(")\n");
	}
	else
	{
		set	= fref / nrate;
		gsc_label_long_comma(set);
		printf(" S/S  (Source = Internal, Alt Ref, Nclk = ");
		gsc_label_long_comma(nclk);
		printf(", Nrate = ");
		gsc_label_long_comma(nrate);
		printf(")\n");
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	ao20_fsamp_ao_report
*
*	Purpose:
*
*		Determine and report the sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		chan	The index of the channel of interest.
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int ao20_fsamp_ao_report(int fd, int index, int verbose, int chan, s32* get)
{
	s32		bit			= 0x1L << chan;
	char	buf[64];
	int		errs		= 0;
	s32		select;
	s32		set			= 0;

	if (verbose)
	{
		sprintf(buf, "Channel %d", chan);
		gsc_label_index(buf, index);
	}

	errs	+= ao20_channel_sel(fd, -1, 0, -1, &select);

	if (select & bit)
	{
		// The channel is enabled.
		errs	+= ao20_fsamp_get(fd, 0, &set);
	}
	else
	{
		// The channel is disabled.
		set	= 0;

		if (verbose)
			printf("0 S/S  (Channel Disabled)\n");
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	ao20_fsamp_ao_report_all
*
*	Purpose:
*
*		Determine and report the sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	Report the overall sample rate here.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int ao20_fsamp_ao_report_all(int fd, int index, int verbose, s32* fsamp)
{
	s32	chans;
	int	chan;
	int	errs;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("Sample Rates", index);

	errs	= ao20_query(fd, -1, 0, AO20_QUERY_CHANNEL_QTY, &chans);

	if (errs == 0)
	{
		if (verbose)
		{
			printf("\n");
			gsc_label_level_inc();
		}

		for (chan = 0; chan < chans; chan++)
		{
			sps		= 0;
			errs	+= ao20_fsamp_ao_report(fd, -1, verbose, chan, &sps);
			total	+= sps;
		}

		if (verbose)
		{
			gsc_label("Overall Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
			gsc_label_level_dec();
		}
	}

	if (fsamp)
		fsamp[0]	= total;

	return(errs);
}


